#### コード2.1, コード2.2

### コード2.1

## オリジナルデータをUCI ML databaseから読み込む
Wine = read.csv(
  url("https://archive.ics.uci.edu/ml/machine-learning-databases/wine/wine.data"), 
  header = FALSE
  )

## 使用する変数(品種, フラボノイド含有量, 色強度)を抽出
wine = Wine[, c(1, 8, 11)]
colnames(wine) = c("grape", "x1", "x2")          # 3変数の命名

## サンプルサイズの計算
# 品種ごと
n1 = sum(wine[, 1] == 1)
n2 = sum(wine[, 1] == 2)
n3 = sum(wine[, 1] == 3)
# トータル
n = n1 + n2 + n3

## 乱数を用いて各品種から教師サンプルを選択(教師サンプルサイズは表2.1の設定)
n1.train = 29; n2.train = 36; n3.train = 24      # 各品種の教師サンプルサイズ
set.seed(380)                                    # 乱数種を指定
# 教師サンプル(品種1)
train1 = sort(sample(which(wine[, 1] == 1), n1.train))
# 教師サンプル(品種2)
train2 = sort(sample(which(wine[, 1] == 2), n2.train))
# 教師サンプル(品種3)
train3 = sort(sample(which(wine[, 1] == 3), n3.train))
# 教師サンプル(全品種)
train  = c(train1, train2, train3)

## 残りをテストサンプルとする
# 各品種のテストサンプルサイズ
n1.test = n1 - n1.train
n2.test = n2 - n2.train
n3.test = n3 - n3.train
# テストサンプル(全品種)
test = (1:n)[-train]

## 教師データ・テストデータへ分割
wine.train = wine[train, ];  wine.test  = wine[test, ]

## 横軸:x1, 縦軸:x2 で散布図作成(図2.1に対応)
r1 = range(wine[, 2]); r2 = range(wine[, 3])     # x1, x2 の値域
par(mfrow = c(1, 2))                             # プロット画面を2分割 

# 教師データの散布図
plot(wine.train[, 2], wine.train[, 3], xlim = r1, ylim = r2,  
     xlab = "x1", ylab = "x2", col = wine.train[, 1], pch = wine.train[, 1],
     main = paste0("training; n1 = ", n1.train, ", n2 = ", n2.train, ", n3 = ", n3.train))
legend("topright", legend = paste0("grape", 1:3), col = 1:3, pch = 1:3) 

# テストデータの散布図 
plot(wine.test[, 2], wine.test[, 3], xlim = r1, ylim = r2,  
     xlab = "x1", ylab = "x2", col = wine.test[, 1], pch = wine.test[, 1],
     main = paste0("test; n1 = ", n1.test, ", n2 = ", n2.test, ", n3 = ", n3.test))
legend("topright", legend = paste0("grape", 1:3), col = 1:3, pch = 1:3) 

par(mfrow = c(1, 1))         # プロット画面の分割を解除 

## 判別境界を作図する際に使用する格子点を作成
grid = data.frame(expand.grid(x1 = seq(r1[1], r1[2], length.out = 500),
                              x2 = seq(r2[1], r2[2], length.out = 500)))

### コード2.2

## 事前確率を設定
pi1 = 1/3; pi2 = 1/3; pi3 = 1- pi1 - pi2

## x1, x2の平均ベクトルを品種ごとに推定 
g = wine.train[, 1]                     # 品種番号の教師データ 
X = as.matrix(wine.train[, 2:3])        # x1, x2の教師データ 
mu1 = colMeans(X[g == 1, ])             # 品種1のx1, x2の平均ベクトル 
mu2 = colMeans(X[g == 2, ])             # 品種2のx1, x2の平均ベクトル 
mu3 = colMeans(X[g == 3, ])             # 品種3のx1, x2の平均ベクトル 
Mu = cbind(mu1, mu2, mu3) 

## x1, x2の分散共分散行列は3品種とも共通と仮定し, これを推定 
# x1, x2の教師データを品種ごとに中心化 
centered.X = t(sapply(1:nrow(X), function(i){X[i, ] - Mu[, g[i]]}))
# 分散共分散行列を推定 
Sigma = var(centered.X)
# 分散共分散行列の逆行列 
Lambda = solve(Sigma)

## 判別ルールを関数化 
LDA = function(X){
   apply(X, 1, function(x){ 
     # 品種1に対する線形判別関数
     d1 = log(pi1 / pi3) + t(x - (mu1 + mu3) / 2) %*% Lambda %*% (mu1 - mu3)
     # 品種2に対する線形判別関数
     d2 = log(pi2 / pi3) + t(x - (mu2 + mu3) / 2) %*% Lambda %*% (mu2 - mu3)
     
     if(d1 > d2 & d1 > 0){   # d1が最大かつ正なら品種1と判別
       return(as.integer(1))
       }else if(d2 > 0){     # d2が最大かつ正なら品種2と判別
         return(as.integer(2))
         }else{              # d1, d2 がいずれも0以下なら品種3と判別
           return(as.integer(3))
           }
     })
}

## 教師データに対する判別結果 
lda.train = LDA(wine.train[, 2:3])                         # 教師データを判別 
print(table(正答 = wine.train[, 1], 判別 = lda.train))     # 正誤表 
lda.err.train = mean(wine.train[, 1] != lda.train)         # 誤判別率 
print(lda.err.train) 

## テストデータに対する判別結果 
lda.test = LDA(wine.test[, 2:3])                           # テストデータを判別 
print(table(正答 = wine.test[, 1], 判別 = lda.test))       # 正誤表 
lda.err.test = mean(wine.test[, 1] != lda.test)            # 誤判別率 
print(lda.err.test)

## コード2.1で作成した格子点gridに対して判別を実行し, 判別境界を作成 
lda.area = LDA(grid)         # 格子点に対して判別を実行 
par(mfrow = c(1, 2))         # プロット画面を2分割 

# 教師データの散布図 
plot(grid, cex = 0.001, col = grey(0.3 + 0.2*lda.area), 
     xlim = r1, ylim = r2) 
par(new = TRUE) 
plot(wine.train[, 2], wine.train[, 3], xlim = r1, ylim = r2,  
     xlab = "", ylab = "", col = wine.train[, 1], pch = wine.train[, 1],
     main = paste0("LDA (training; n1 = ", n1.train,  
                   ", n2 = ", n2.train, ", n3 = ", n3.train, ")  \n",  
                   "error rate = ", signif(lda.err.train, 4))) 
legend("topright", legend = paste0("grape", 1:3), col = 1:3, pch = 1:3) 

# テストデータの散布図 
plot(grid, cex = 0.001, col = grey(0.3 + 0.2*lda.area), 
     xlim = r1, ylim = r2) 
par(new = TRUE) 
plot(wine.test[, 2], wine.test[, 3], xlim = r1, ylim = r2,  
     xlab = "", ylab = "", col = wine.test[, 1], pch = wine.test[, 1],
     main = paste0("LDA (test; n1 = ", n1.test,  
                   ", n2 = ", n2.test, ", n3 = ", n3.test, ") \n",  
                   "error rate = ", signif(lda.err.test, 4))) 
legend("topright", legend = paste0("grape", 1:3), col = 1:3, pch = 1:3) 

par(mfrow = c(1, 1))         # プロット画面の分割を解除 
