# pythonの準備とデータの取得
rm(list=ls())
Sys.setenv(TENSORFLOW_PYTHON = "/anaconda3/bin/python3.6")

library(tensorflow)
library(keras)

cifar10 <- dataset_cifar10()
c(c(x_train, y_train), c(x_test, y_test)) %<-% cifar10

# クラス6 (カエル)の画像のみ取り出す
x_train <- x_train[as.integer(y_train) == 6,,,]
x_train <- x_train / 255

# 生成器の定義(コード3.13)
latent_dim <- 32
channels <- 3

generator_input <- layer_input(shape = c(latent_dim))
generator_output <- generator_input %>%
  layer_dense(units = 128 * 16 * 16) %>%
  layer_activation_leaky_relu() %>%
  layer_reshape(target_shape = c(16, 16, 128)) %>%
  layer_conv_2d(filters = 256, kernel_size = 5, padding = "same") %>%
  layer_activation_leaky_relu() %>%
  layer_conv_2d_transpose(filters = 256, kernel_size = 4, strides = 2, padding = "same") %>%
  layer_activation_leaky_relu() %>%
  layer_conv_2d(filters = 256, kernel_size = 5, padding = "same") %>%
  layer_activation_leaky_relu() %>%
  layer_conv_2d(filters = 256, kernel_size = 5, padding = "same") %>%
  layer_activation_leaky_relu() %>%
  layer_conv_2d(filters = channels, kernel_size = 7, activation = "tanh", padding = "same")

generator <- keras_model(generator_input, generator_output)
generator

# 分類器の定義(コード3.14)
height <- 32
width <- 32

discriminator_input <- layer_input(shape = c(height, width, channels))
discriminator_output <- discriminator_input %>%
  layer_conv_2d(filters = 128, kernel_size = 3) %>%
  layer_activation_leaky_relu() %>%
  layer_conv_2d(filters = 128, kernel_size = 4, strides = 2) %>%
  layer_activation_leaky_relu() %>%
  layer_conv_2d(filters = 128, kernel_size = 4, strides = 2) %>%
  layer_activation_leaky_relu() %>%
  layer_conv_2d(filters = 128, kernel_size = 4, strides = 2) %>%
  layer_activation_leaky_relu() %>%
  layer_flatten() %>%
  layer_dropout(rate = 0.4) %>%
  layer_dense(units = 1, activation = "sigmoid")

discriminator <- keras_model(discriminator_input, discriminator_output)
discriminator

discriminator_optimizer <- optimizer_rmsprop(
  lr = 0.0008,
  clipvalue = 1.0,
  decay = 1e-8
)

discriminator %>% compile(
  optimizer = discriminator_optimizer,
  loss = "binary_crossentropy"
)

# ganの定義
freeze_weights(discriminator)
gan_input <- layer_input(shape = c(latent_dim))
gan_output <- discriminator(generator(gan_input))
gan <- keras_model(gan_input, gan_output)
gan_optimizer <- optimizer_rmsprop(
  lr = 0.0004,
  clipvalue = 1.0,
  decay = 1e-8
)

gan %>% compile(
  optimizer = gan_optimizer,
  loss = "binary_crossentropy"
)
gan

# ganのパラメータ推定
iterations <- 10000
batch_size <- 20
start <- 1

for (step in 1:iterations) {
  # 生成器による偽の画像をミニバッチ数だけ生成
  random_latent_vectors <- matrix(rnorm(batch_size * latent_dim), nrow = batch_size, ncol = latent_dim) 
  generated_images <- generator %>% predict(random_latent_vectors)

  # 実際のデータと偽のデータを統合(真の画像のラベルが0)
  stop <- start + batch_size - 1
  real_images <- x_train[start:stop,,,]
  rows <- nrow(real_images)
  combined_images <- array(0, dim = c(rows * 2, dim(real_images)[-1]))
  combined_images[1:rows,,,] <- generated_images
  combined_images[(rows+1):(rows*2),,,] <- real_images
  labels <- rbind(matrix(1, nrow = batch_size, ncol = 1), matrix(0, nrow = batch_size, ncol = 1))
  ## 一様ノイズによるラベルの無作為化 
  ### ※ データを無作為化して推定を安定させるためのトリック
  labels <- labels + (0.5 * array(runif(prod(dim(labels))), dim = dim(labels)))
  
  # 分類器の更新
  d_loss <- discriminator %>% train_on_batch(combined_images, labels) 
  
  random_latent_vectors <- matrix(rnorm(batch_size * latent_dim), nrow = batch_size, ncol = latent_dim)
  misleading_targets <- array(0, dim = c(batch_size, 1))
  
  # 生成器(敵対的損失)の更新
  a_loss <- gan %>% train_on_batch(
    random_latent_vectors,
    misleading_targets
  )
  
  start <- start + batch_size
  if (start > (nrow(x_train) - batch_size))
    start <- 1
  if (step %% 100 == 0) {
    # 分類損失と敵対的損失の表示
    # 必要なら, 以下のコマンドでモデルのパラメータも保存できる
    # save_model_weights_hdf5(gan, "gan.h5")
    cat("discriminator loss:", d_loss, "\n")
    cat("adversarial loss:", a_loss, "\n")
    
    # 100ステップごとに生成された画像と実際の画像を出力する
    # paste0で保存先のパスを指定しておくと良い
    # 以下は, デスクトップにあるGANというフォルダに保存するときの例
    image_array_save(
      generated_images[1,,,] * 255,
      path = file.path(paste0("./Desktop/GAN/generated_frog", step, ".png"))
    )
    image_array_save(
      real_images[1,,,] * 255,
      path = file.path(paste0("./Desktop/GAN/real_frog", step, ".png"))
    )
  }
}
