# pythonの準備とデータの取得
Sys.setenv(TENSORFLOW_PYTHON = "/anaconda3/bin/python3.6")

library(tensorflow)
library(keras)
library(R6)

mnist <- dataset_mnist()
x_train <- mnist$train$x
x_test <- mnist$test$x

# 入力を区間[0,1]に値をとるようスケーリング(入力は28*28行列)
# 正解ラベルは使用しない
x_train <- array_reshape(x_train, c(dim(x_train), 1)) / 255
x_test <- array_reshape(x_test, c(dim(x_test), 1)) / 255

# 符号化器の定義(コード3.9)
q <- 2L
input <- layer_input(shape = c(28, 28, 1))
x <- input %>%
  layer_conv_2d(filters = 64, kernel_size = 3, padding = "same", activation = "relu") %>%
  layer_conv_2d(filters = 64, kernel_size = 3, padding = "same", activation = "relu") %>%
  layer_max_pooling_2d(pool_size = c(2, 2)) %>%
  layer_conv_2d(filters = 32, kernel_size = 3, padding = "same", activation = "relu") %>%
  layer_conv_2d(filters = 32, kernel_size = 3, padding = "same", activation = "relu") %>%
  layer_max_pooling_2d(pool_size = c(2, 2))

dim_convolution <- k_int_shape(x)

x <- x %>%
  layer_flatten() %>%
  layer_dense(units = 32, activation = "relu")

z_mean <- x %>% layer_dense(units = q)
z_log_var <- x %>% layer_dense(units = q)

encoder_latent <- keras_model(input, z_mean)
encoder_latent


# 潜在変数のサンプリング(コード3.10)
sampling <- function(args) {
  c(z_mean, z_log_var) %<-% args
  epsilon <- k_random_normal(shape = list(k_shape(z_mean)[1], q), mean = 0, stddev = 1)
  z_mean + k_exp(z_log_var) * epsilon
}

z <- list(z_mean, z_log_var) %>% layer_lambda(sampling)

# 復号化器の定義(コード3.11)
decoder_input<- layer_input(k_int_shape(z)[-1])
x <- decoder_input %>%
  layer_dense(units = prod(as.integer(dim_convolution[-1])), activation = "relu") %>%
  layer_reshape(target_shape = dim_convolution[-1]) %>%
  layer_conv_2d_transpose(filters = 32, kernel_size = 3, padding = "same", activation = "relu", stride = 2) %>%
  layer_conv_2d(filters = 32, kernel_size = 3, padding = "same", activation = "relu") %>%
  layer_conv_2d_transpose(filters = 64, kernel_size = 3, padding = "same", activation = "relu", stride = 2) %>%
  layer_conv_2d(filters = 64, kernel_size = 3, padding = "same", activation = "relu") %>%
  layer_conv_2d(filters = 1, kernel_size = 3, padding = "same", activation = "sigmoid")

decoder <- keras_model(decoder_input, x)
decoder

# 誤差関数の定義とモデルの当てはめ(コード3.12)
z_decoded <- decoder(z)
CustomVariationalLayer <- R6Class("CustomVariationalLayer",
  inherit = KerasLayer,
  public = list(
    vae_loss = function(x, z_decoded) {
      x <- k_flatten(x)
      z_decoded <- k_flatten(z_decoded)
      xent_loss <- metric_binary_crossentropy(x, z_decoded)
      kl_loss <- -5e-4 * k_mean(
        1 + z_log_var - k_square(z_mean) - k_exp(z_log_var),
        axis = -1L
      )
      k_mean(xent_loss + kl_loss)
    },
    call = function(inputs, mask = NULL) {
      x <- inputs[[1]]
      z_decoded <- inputs[[2]]
      loss <- self$vae_loss(x, z_decoded)
      self$add_loss(loss, inputs = inputs)
      x
    }
  )
)

layer_variational <- function(object) {
  create_layer(CustomVariationalLayer, object, list())
}
y <- list(input, z_decoded) %>% layer_variational()

vae <- keras_model(input, y)
vae

vae %>% compile(
  optimizer = "rmsprop",
  loss = NULL
)

history<- vae %>% fit(
  x = x_train, y = NULL,
  epochs = 10,
  batch_size = 128,
  validation_data = list(x_test, NULL)
)


##### 潜在空間と復号化器の可視化
# テストデータの正解ラベル
y_test <- mnist$test$y

# 符号化された(潜在空間での)座標
x_encoded <- encoder_latent %>% predict(x_test, batch_size = 128)

mycol<- sapply(1:length(y_test), function(i){rainbow(10)[y_test[i]+1]})
center<- sapply(0:9, function(i){colMeans(x_encoded[y_test==i, ])})

plot(x_encoded, col=mycol, bg=NULL, pch=16, xlab="z1", ylab="z2", cex.axis=1.5, cex.lab=1.5)
grid()
for(i in 1:10){text(center[1, i], center[2, i], i-1, cex=2)}

# 潜在空間から生成される画像
## 15*15のラスター画像で表示
n <- 15
digit_size <- 28
grid_x <- seq(-15, 15, length = n)
grid_y <- seq(-15, 15, length = n)
op <- par(mfrow = c(n, n), mar = c(0,0,0,0), bg = "black")
for (i in rev(1:length(grid_x))) {
  yi <- grid_x[[i]]
  for (j in 1:length(grid_y)) {
    xi <- grid_y[[j]]
    z_sample <- matrix(c(xi, yi), nrow = 1, ncol = 2)
    z_sample <- t(replicate(128, z_sample, simplify = "matrix"))
    x_decoded <- decoder %>% predict(z_sample, 128)
    digit <- array_reshape(x_decoded[1,,,], dim = c(digit_size, digit_size))
    plot(as.raster(digit))
  }
}
par(op)



